@@ -77,11 +77,10 @@ int packet_decrypt(ssh_session session, void *data,uint32_t len) {
     SAFE_FREE(out);
     return -1;
   }
-  crypto->cbc_decrypt(crypto,data,out,len);
+  crypto->decrypt(crypto,data,out,len);
 
   memcpy(data,out,len);
-  memset(out,0,len);
-
+  BURN_BUFFER(out, len);
   SAFE_FREE(out);
   return 0;
 }
@@ -92,6 +91,7 @@ unsigned char *packet_encrypt(ssh_session session, void *data, uint32_t len) {
   char *out = NULL;
   unsigned int finallen;
   uint32_t seq;
+  enum ssh_hmac_e type;
 
   assert(len);
 
@@ -107,6 +107,7 @@ unsigned char *packet_encrypt(ssh_session session, void *data, uint32_t len) {
     return NULL;
   }
 
+  type = session->current_crypto->out_hmac;
   seq = ntohl(session->send_seq);
   crypto = session->current_crypto->out_cipher;
 
@@ -117,7 +118,7 @@ unsigned char *packet_encrypt(ssh_session session, void *data, uint32_t len) {
   }
 
   if (session->version == 2) {
-    ctx = hmac_init(session->current_crypto->encryptMAC,20,SSH_HMAC_SHA1);
+    ctx = hmac_init(session->current_crypto->encryptMAC, hmac_digest_len(type), type);
     if (ctx == NULL) {
       SAFE_FREE(out);
       return NULL;
@@ -126,18 +127,18 @@ unsigned char *packet_encrypt(ssh_session session, void *data, uint32_t len) {
     hmac_update(ctx,data,len);
     hmac_final(ctx,session->current_crypto->hmacbuf,&finallen);
 #ifdef DEBUG_CRYPTO
-    ssh_print_hexa("mac: ",data,len);
-    if (finallen != 20) {
+    ssh_print_hexa("mac: ",data,hmac_digest_len(type));
+    if (finallen != hmac_digest_len(type)) {
       printf("Final len is %d\n",finallen);
     }
-    ssh_print_hexa("Packet hmac", session->current_crypto->hmacbuf, 20);
+    ssh_print_hexa("Packet hmac", session->current_crypto->hmacbuf, hmac_digest_len(type));
 #endif
   }
 
-  crypto->cbc_encrypt(crypto, data, out, len);
+  crypto->encrypt(crypto, data, out, len);
 
   memcpy(data, out, len);
-  memset(out, 0, len);
+  BURN_BUFFER(out, len);
   SAFE_FREE(out);
 
   if (session->version == 2) {
@@ -160,13 +161,13 @@ unsigned char *packet_encrypt(ssh_session session, void *data, uint32_t len) {
  *                      occurred.
  */
 int packet_hmac_verify(ssh_session session, ssh_buffer buffer,
-    unsigned char *mac) {
-  unsigned char hmacbuf[EVP_MAX_MD_SIZE] = {0};
+    unsigned char *mac, enum ssh_hmac_e type) {
+  unsigned char hmacbuf[DIGEST_MAX_LEN] = {0};
   HMACCTX ctx;
   unsigned int len;
   uint32_t seq;
 
-  ctx = hmac_init(session->current_crypto->decryptMAC, 20, SSH_HMAC_SHA1);
+  ctx = hmac_init(session->current_crypto->decryptMAC, hmac_digest_len(type), type);
   if (ctx == NULL) {
     return -1;
   }
