@@ -42,7 +42,7 @@ sftp_client_message sftp_get_client_message(sftp_session sftp) {
   sftp_packet packet;
   sftp_client_message msg;
   ssh_buffer payload;
-  ssh_string tmp;
+  int rc;
 
   msg = malloc(sizeof (struct sftp_client_message_struct));
   if (msg == NULL) {
@@ -64,7 +64,9 @@ sftp_client_message sftp_get_client_message(sftp_session sftp) {
 
   /* take a copy of the whole packet */
   msg->complete_message = ssh_buffer_new();
-  buffer_add_data(msg->complete_message, buffer_get_rest(payload), buffer_get_rest_len(payload));
+  ssh_buffer_add_data(msg->complete_message,
+                      buffer_get_rest(payload),
+                      buffer_get_rest_len(payload));
 
   buffer_get_u32(payload, &msg->id);
 
@@ -79,25 +81,24 @@ sftp_client_message sftp_get_client_message(sftp_session sftp) {
       }
       break;
     case SSH_FXP_READ:
-      msg->handle = buffer_get_ssh_string(payload);
-      if (msg->handle == NULL) {
+      rc = ssh_buffer_unpack(payload,
+                             "Sqd",
+                             &msg->handle,
+                             &msg->offset,
+                             &msg->len);
+      if (rc != SSH_OK) {
         ssh_set_error_oom(session);
         sftp_client_message_free(msg);
         return NULL;
       }
-      buffer_get_u64(payload, &msg->offset);
-      buffer_get_u32(payload, &msg->len);
       break;
     case SSH_FXP_WRITE:
-      msg->handle = buffer_get_ssh_string(payload);
-      if (msg->handle == NULL) {
-        ssh_set_error_oom(session);
-        sftp_client_message_free(msg);
-        return NULL;
-      }
-      buffer_get_u64(payload, &msg->offset);
-      msg->data = buffer_get_ssh_string(payload);
-      if (msg->data == NULL) {
+      rc = ssh_buffer_unpack(payload,
+                             "SqS",
+                             &msg->handle,
+                             &msg->offset,
+                             &msg->data);
+      if (rc != SSH_OK) {
         ssh_set_error_oom(session);
         sftp_client_message_free(msg);
         return NULL;
@@ -108,15 +109,10 @@ sftp_client_message sftp_get_client_message(sftp_session sftp) {
     case SSH_FXP_OPENDIR:
     case SSH_FXP_READLINK:
     case SSH_FXP_REALPATH:
-      tmp = buffer_get_ssh_string(payload);
-      if (tmp == NULL) {
-        ssh_set_error_oom(session);
-        sftp_client_message_free(msg);
-        return NULL;
-      }
-      msg->filename = ssh_string_to_char(tmp);
-      ssh_string_free(tmp);
-      if (msg->filename == NULL) {
+      rc = ssh_buffer_unpack(payload,
+                             "s",
+                             &msg->filename);
+      if (rc != SSH_OK) {
         ssh_set_error_oom(session);
         sftp_client_message_free(msg);
         return NULL;
@@ -124,21 +120,11 @@ sftp_client_message sftp_get_client_message(sftp_session sftp) {
       break;
     case SSH_FXP_RENAME:
     case SSH_FXP_SYMLINK:
-      tmp = buffer_get_ssh_string(payload);
-      if (tmp == NULL) {
-        ssh_set_error_oom(session);
-        sftp_client_message_free(msg);
-        return NULL;
-      }
-      msg->filename = ssh_string_to_char(tmp);
-      ssh_string_free(tmp);
-      if (msg->filename == NULL) {
-        ssh_set_error_oom(session);
-        sftp_client_message_free(msg);
-        return NULL;
-      }
-      msg->data = buffer_get_ssh_string(payload);
-      if (msg->data == NULL) {
+      rc = ssh_buffer_unpack(payload,
+                             "sS",
+                             &msg->filename,
+                             &msg->data);
+      if (rc != SSH_OK) {
         ssh_set_error_oom(session);
         sftp_client_message_free(msg);
         return NULL;
@@ -146,15 +132,10 @@ sftp_client_message sftp_get_client_message(sftp_session sftp) {
       break;
     case SSH_FXP_MKDIR:
     case SSH_FXP_SETSTAT:
-      tmp = buffer_get_ssh_string(payload);
-      if (tmp == NULL) {
-        ssh_set_error_oom(session);
-        sftp_client_message_free(msg);
-        return NULL;
-      }
-      msg->filename=ssh_string_to_char(tmp);
-      ssh_string_free(tmp);
-      if (msg->filename == NULL) {
+      rc = ssh_buffer_unpack(payload,
+                             "s",
+                             &msg->filename);
+      if (rc != SSH_OK) {
         ssh_set_error_oom(session);
         sftp_client_message_free(msg);
         return NULL;
@@ -182,38 +163,28 @@ sftp_client_message sftp_get_client_message(sftp_session sftp) {
       break;
     case SSH_FXP_LSTAT:
     case SSH_FXP_STAT:
-      tmp = buffer_get_ssh_string(payload);
-      if (tmp == NULL) {
-        ssh_set_error_oom(session);
-        sftp_client_message_free(msg);
-        return NULL;
-      }
-      msg->filename = ssh_string_to_char(tmp);
-      ssh_string_free(tmp);
-      if (msg->filename == NULL) {
+      rc = ssh_buffer_unpack(payload,
+                             "s",
+                             &msg->filename);
+      if (rc != SSH_OK) {
         ssh_set_error_oom(session);
         sftp_client_message_free(msg);
         return NULL;
       }
       if(sftp->version > 3) {
-        buffer_get_u32(payload,&msg->flags);
+        ssh_buffer_unpack(payload, "d", &msg->flags);
       }
       break;
     case SSH_FXP_OPEN:
-      tmp=buffer_get_ssh_string(payload);
-      if (tmp == NULL) {
+      rc = ssh_buffer_unpack(payload,
+                             "sd",
+                             &msg->filename,
+                             &msg->flags);
+      if (rc != SSH_OK) {
         ssh_set_error_oom(session);
         sftp_client_message_free(msg);
         return NULL;
       }
-      msg->filename = ssh_string_to_char(tmp);
-      ssh_string_free(tmp);
-      if (msg->filename == NULL) {
-        ssh_set_error_oom(session);
-        sftp_client_message_free(msg);
-        return NULL;
-      }
-      buffer_get_u32(payload,&msg->flags);
       msg->attr = sftp_parse_attr(sftp, payload, 0);
       if (msg->attr == NULL) {
         ssh_set_error_oom(session);
@@ -222,13 +193,15 @@ sftp_client_message sftp_get_client_message(sftp_session sftp) {
       }
       break;
     case SSH_FXP_FSTAT:
-      msg->handle = buffer_get_ssh_string(payload);
-      if (msg->handle == NULL) {
+      rc = ssh_buffer_unpack(payload,
+                             "Sd",
+                             &msg->handle,
+                             &msg->flags);
+      if (rc != SSH_OK) {
         ssh_set_error_oom(session);
         sftp_client_message_free(msg);
         return NULL;
       }
-      buffer_get_u32(payload, &msg->flags);
       break;
     default:
       ssh_set_error(sftp->session, SSH_FATAL,
@@ -237,9 +210,6 @@ sftp_client_message sftp_get_client_message(sftp_session sftp) {
       return NULL;
   }
 
-  msg->flags = ntohl(msg->flags);
-  msg->offset = ntohll(msg->offset);
-  msg->len = ntohl(msg->len);
   sftp_packet_free(packet);
 
   return msg;
@@ -407,7 +377,7 @@ int sftp_reply_names(sftp_client_message msg) {
 
   if (buffer_add_u32(out, msg->id) < 0 ||
       buffer_add_u32(out, htonl(msg->attr_num)) < 0 ||
-      buffer_add_data(out, buffer_get_rest(msg->attrbuf),
+      ssh_buffer_add_data(out, buffer_get_rest(msg->attrbuf),
         buffer_get_rest_len(msg->attrbuf)) < 0 ||
       sftp_packet_write(msg->sftp, SSH_FXP_NAME, out) < 0) {
     ssh_buffer_free(out);
@@ -466,7 +436,7 @@ int sftp_reply_data(sftp_client_message msg, const void *data, int len) {
 
   if (buffer_add_u32(out, msg->id) < 0 ||
       buffer_add_u32(out, ntohl(len)) < 0 ||
-      buffer_add_data(out, data, len) < 0 ||
+      ssh_buffer_add_data(out, data, len) < 0 ||
       sftp_packet_write(msg->sftp, SSH_FXP_DATA, out) < 0) {
     ssh_buffer_free(out);
     return -1;
