@@ -20,7 +20,7 @@ static void setup(void **state) {
     user = getenv("TORTURE_USER");
     password = getenv("TORTURE_PASSWORD");
 
-    session = torture_ssh_session(host, user, password);
+    session = torture_ssh_session(host, NULL, user, password);
     assert_false(session == NULL);
     t = torture_sftp_session(session);
     assert_false(t == NULL);
@@ -45,12 +45,14 @@ static void torture_sftp_read_blocking(void **state) {
     ssize_t byteswritten;
     int fd;
     sftp_file file;
-
+    mode_t mask;
 
     file = sftp_open(t->sftp, "/usr/bin/ssh", O_RDONLY, 0);
     assert_non_null(file);
 
+    mask = umask(S_IRWXO | S_IRWXG);
     fd = mkstemp(libssh_tmp_file);
+    umask(mask);
     unlink(libssh_tmp_file);
 
     for (;;) {
@@ -70,12 +72,13 @@ static void torture_sftp_read_blocking(void **state) {
 
 int torture_run_tests(void) {
     int rc;
-    const UnitTest tests[] = {
+    UnitTest tests[] = {
         unit_test_setup_teardown(torture_sftp_read_blocking, setup, teardown)
     };
 
     ssh_init();
 
+    torture_filter_tests(tests);
     rc = run_tests(tests);
     ssh_finalize();
 
