@@ -37,11 +37,12 @@ static void torture_channel_select(void **state)
 
 int torture_run_tests(void) {
     int rc;
-    const UnitTest tests[] = {
+    UnitTest tests[] = {
         unit_test(torture_channel_select),
     };
 
     ssh_init();
+    torture_filter_tests(tests);
     rc = run_tests(tests);
     ssh_finalize();
 
